@@ -9,7 +9,7 @@
  * From http://home.arcor.de/ahuggel/exiv2/
  */
 
-/* Copyright 2005-2023 Daniel Foote, Dan Fandrich.
+/* Copyright 2005-2025 Daniel Foote, Dan Fandrich.
  *
  * This file is part of gpscorrelate.
  *
@@ -85,7 +85,7 @@ int main(int argc, char* argv[])
 	printf("Done write, now reading...\n");
 
 	int GPS = 0;
-	char* Ret = ReadExifDate(argv[1], &GPS);
+	char* Ret = ReadExifData(argv[1], &GPS, NULL, NULL, NULL, NULL);
 	if (Ret)
 	{
 		printf("Date: %s.\n", Ret);
@@ -103,62 +103,11 @@ int main(int argc, char* argv[])
 };
 */
 
-char* ReadExifDate(const char* File, int* IncludesGPS)
+char* ReadExifData(const char* File, int* IncludesGPS, double* Lat, double* Long, double* Elev,
+		long *OffsetTime)
 {
-	// Open and read the file.
-	Exiv2::Image::UNIQUEPTR Image;
+	*IncludesGPS = 0;
 
-	try {
-		Image = Exiv2::ImageFactory::open(File);
-	} catch (Exiv2::Error& e) {
-		DEBUGLOG("Failed to open file %s.\n", File);
-		return NULL;
-	}
-	Image->readMetadata();
-	if (Image.get() == NULL)
-	{
-		DEBUGLOG("Failed to read file %s %s.\n",
-			 File, Exiv2::strError().c_str());
-		return NULL;
-	}
-
-	Exiv2::ExifData &ExifRead = Image->exifData();
-
-	// Read the tag out.
-	Exiv2::Exifdatum& Tag = ExifRead["Exif.Photo.DateTimeOriginal"];
-
-	// Check that the tag is not blank.
-	std::string Value = Tag.toString();
-
-	if (Value.length() == 0)
-	{
-		// No date/time stamp.
-		// Not good.
-		// Just return - above us will handle it.
-		return NULL;
-	}
-
-	// Copy the tag and return that.
-	char* Copy = strdup(Value.c_str());
-	
-	// Check if we have GPS tags.
-	Exiv2::Exifdatum& GPSData = ExifRead["Exif.GPSInfo.GPSLatitude"];
-
-	if (GPSData.count() < 3)
-	{
-		// No valid GPS data.
-		*IncludesGPS = 0;
-	} else {
-		// Seems to include GPS data...
-		*IncludesGPS = 1;
-	}
-
-	// Now return, passing a pointer to the date string.
-	return Copy; // It's up to the caller to free this.
-}
-
-char* ReadExifData(const char* File, double* Lat, double* Long, double* Elev, int* IncludesGPS)
-{
 	// This function varies in that it reads
 	// much more data than the last, specifically
 	// for display purposes. For the GUI version.
@@ -171,14 +120,20 @@ char* ReadExifData(const char* File, double* Lat, double* Long, double* Elev, in
 		DEBUGLOG("Failed to open file %s.\n", File);
 		return NULL;
 	}
-	Image->readMetadata();
 	if (Image.get() == NULL)
 	{
 		DEBUGLOG("Failed to read file %s %s.\n",
 			 File, Exiv2::strError().c_str());
 		return NULL;
 	}
-	
+
+	try {
+		Image->readMetadata();
+	} catch (Exiv2::Error& e) {
+		DEBUGLOG("Failed to read EXIF metadata in %s.\n", File);
+		return NULL;
+	}
+
 	Exiv2::ExifData &ExifRead = Image->exifData();
 
 	// Read the tag out.
@@ -195,111 +150,141 @@ char* ReadExifData(const char* File, double* Lat, double* Long, double* Elev, in
 		return NULL;
 	}
 
+	// Look for subseconds
+	Exiv2::Exifdatum& SubsecTag = ExifRead["Exif.Photo.SubSecTimeOriginal"];
+	std::string SubsecValue = SubsecTag.toString();
+	if (SubsecValue.length()) {
+		// Append the subseconds to the time string
+		Value.append(".");
+		Value.append(SubsecValue);
+	}
+
 	// Copy the tag and return that.
-	char* Copy = strdup(Value.c_str());
-	
+	char* DateTime = strdup(Value.c_str());
+
+	// Look for time offset
+	if (OffsetTime) {
+#if EXIV2_TEST_VERSION(0, 27, 4)
+		Exiv2::Exifdatum& OffsetTimeTag = ExifRead["Exif.Photo.OffsetTimeOriginal"];
+		std::string OffsetTimeValue = OffsetTimeTag.toString();
+		if (OffsetTimeValue.length() == 6) {
+			/* Split into hours, minutes with the same sign for both. */
+			int OffsetHours, OffsetMins;
+			if (sscanf(OffsetTimeValue.c_str(), "%d:%d", &OffsetHours, &OffsetMins) == 2) {
+				/* Can't look at sign of hours because it might be -0 */
+				if (OffsetTimeValue[0] == '-')
+					OffsetMins *= -1;
+				*OffsetTime = OffsetHours * 3600 + OffsetMins * 60;
+			} else
+				*OffsetTime = NO_OFFSET_TIME;
+		} else
+#endif
+			*OffsetTime = NO_OFFSET_TIME;
+	}
+
 	// Check if we have GPS tags.
-	*IncludesGPS = 0;
 	Exiv2::Exifdatum GPSData = ExifRead["Exif.GPSInfo.GPSVersionID"];
 
 	Value = GPSData.toString();
 
-	if (Value.length() == 0)
-	{
-		// No GPS version tag.
-		DEBUGLOG("No GPSVersionID.\n");
-
-		// The EXIF spec says that this tag is mandatory if the GPS IFD
-		// exists, but some cameras skip it anyway.
-	}
-
-	// Read it out and send it up!
-	// What we are trying to do here is convert the
-	// three rationals:
-	//    dd/v mm/v ss/v
-	// To a decimal
-	//    dd.dddddd...
-	// dd/v is easy: result = dd/v.
-	// mm/v is harder:
-	//    mm
-	//    -- / 60 = result.
-	//     v
-	// ss/v is sorta easy.
-	//     ss   
-	//     -- / 3600 = result
-	//      v   
-	// Each part is added to the final number.
-	Exiv2::URational RatNum;
-
 	GPSData = ExifRead["Exif.GPSInfo.GPSLatitude"];
-	if (GPSData.count() < 3)
-		*Lat = nan("invalid");
-	else {
-		// This is enough to say it includes GPS data...
-		*IncludesGPS = 1;
+	*IncludesGPS = GPSData.count() == 3;
 
-		RatNum = GPSData.toRational(0);
-		*Lat = (double)RatNum.first / (double)RatNum.second;
-		RatNum = GPSData.toRational(1);
-		*Lat = *Lat + (((double)RatNum.first / (double)RatNum.second) / 60);
-		RatNum = GPSData.toRational(2);
-		*Lat = *Lat + (((double)RatNum.first / (double)RatNum.second) / 3600);
-
-		GPSData = ExifRead["Exif.GPSInfo.GPSLatitudeRef"];
-		if (strcmp(GPSData.toString().c_str(), "S") == 0)
+	// Skip Lat/Long/Elev reading if the caller doesn't need it
+	if (Lat) {
+		if (Value.length() == 0)
 		{
-			// Negate the value - Western Hemisphere.
-			*Lat = -*Lat;
+			// No GPS version tag.
+			DEBUGLOG("No GPSVersionID.\n");
+
+			// The EXIF spec says that this tag is mandatory if the GPS IFD
+			// exists, but some cameras skip it anyway.
 		}
-	}
-	
-	GPSData = ExifRead["Exif.GPSInfo.GPSLongitude"];
-	if (GPSData.count() < 3)
-		*Long = nan("invalid");
-	else {
-		RatNum = GPSData.toRational(0);
-		*Long = (double)RatNum.first / (double)RatNum.second;
-		RatNum = GPSData.toRational(1);
-		*Long = *Long + (((double)RatNum.first / (double)RatNum.second) / 60);
-		RatNum = GPSData.toRational(2);
-		*Long = *Long + (((double)RatNum.first / (double)RatNum.second) / 3600);
 
-		GPSData = ExifRead["Exif.GPSInfo.GPSLongitudeRef"];
-		if (strcmp(GPSData.toString().c_str(), "W") == 0)
-		{
-			// Negate the value - Western Hemisphere.
-			*Long = -*Long;
+		// Read it out and send it up!
+		// What we are trying to do here is convert the
+		// three rationals:
+		//	  dd/v mm/v ss/v
+		// To a decimal
+		//	  dd.dddddd...
+		// dd/v is easy: result = dd/v.
+		// mm/v is harder:
+		//	  mm
+		//	  -- / 60 = result.
+		//	   v
+		// ss/v is sorta easy.
+		//	   ss
+		//	   -- / 3600 = result
+		//		v
+		// Each part is added to the final number.
+		Exiv2::URational RatNum;
+
+		if (GPSData.count() != 3)
+			*Lat = nan("invalid");
+		else {
+			// This is enough to say it includes GPS data...
+			*IncludesGPS = 1;
+
+			RatNum = GPSData.toRational(0);
+			*Lat = (double)RatNum.first / (double)RatNum.second;
+			RatNum = GPSData.toRational(1);
+			*Lat = *Lat + (((double)RatNum.first / (double)RatNum.second) / 60);
+			RatNum = GPSData.toRational(2);
+			*Lat = *Lat + (((double)RatNum.first / (double)RatNum.second) / 3600);
+
+			GPSData = ExifRead["Exif.GPSInfo.GPSLatitudeRef"];
+			if (strcmp(GPSData.toString().c_str(), "S") == 0)
+			{
+				// Negate the value - Western Hemisphere.
+				*Lat = -*Lat;
+			}
 		}
-	}
 
-	// Finally, read elevation out. This one is simple.
-	GPSData = ExifRead["Exif.GPSInfo.GPSAltitude"];
-	if (GPSData.count() < 1)
-		*Elev = nan("invalid");
-	else {
-		RatNum = GPSData.toRational(0);
-		*Elev = (double)RatNum.first / (double)RatNum.second;
+		GPSData = ExifRead["Exif.GPSInfo.GPSLongitude"];
+		if (GPSData.count() != 3)
+			*Long = nan("invalid");
+		else {
+			RatNum = GPSData.toRational(0);
+			*Long = (double)RatNum.first / (double)RatNum.second;
+			RatNum = GPSData.toRational(1);
+			*Long = *Long + (((double)RatNum.first / (double)RatNum.second) / 60);
+			RatNum = GPSData.toRational(2);
+			*Long = *Long + (((double)RatNum.first / (double)RatNum.second) / 3600);
 
-		// Is the altitude below sea level? If so, negate the value.
-		GPSData = ExifRead["Exif.GPSInfo.GPSAltitudeRef"];
-		if (GPSData.count() >= 1 && (int) GPSData.TOINTEGER() == 1)
-		{
-				// Negate the elevation.
-				*Elev = -*Elev;
+			GPSData = ExifRead["Exif.GPSInfo.GPSLongitudeRef"];
+			if (strcmp(GPSData.toString().c_str(), "W") == 0)
+			{
+				// Negate the value - Western Hemisphere.
+				*Long = -*Long;
+			}
+		}
+
+		// Finally, read elevation out. This one is simple.
+		GPSData = ExifRead["Exif.GPSInfo.GPSAltitude"];
+		if (GPSData.count() != 1)
+			*Elev = nan("invalid");
+		else {
+			RatNum = GPSData.toRational(0);
+			*Elev = (double)RatNum.first / (double)RatNum.second;
+
+			// Is the altitude below sea level? If so, negate the value.
+			GPSData = ExifRead["Exif.GPSInfo.GPSAltitudeRef"];
+			if (GPSData.count() >= 1 && (int) GPSData.TOINTEGER() == 1)
+			{
+					// Negate the elevation.
+					*Elev = -*Elev;
+			}
 		}
 	}
 
 	// Now return, passing a pointer to the date string.
-	return Copy; // It's up to the caller to free this.
+	return DateTime; // It's up to the caller to free this.
 }
 
 // This function is for the --fix-datestamp option.
 // DateStamp and TimeStamp should be 12-char strings.
 char* ReadGPSTimestamp(const char* File, char* DateStamp, char* TimeStamp, int* IncludesGPS)
 {
-	// This function varies in that it reads
-	// much more data than the last, specifically
-	// for display purposes. For the GUI version.
 	// Open and read the file.
 	Exiv2::Image::UNIQUEPTR Image;
 
@@ -309,14 +294,20 @@ char* ReadGPSTimestamp(const char* File, char* DateStamp, char* TimeStamp, int*
 		DEBUGLOG("Failed to open file %s.\n", File);
 		return NULL;
 	}
-	Image->readMetadata();
 	if (Image.get() == NULL)
 	{
 		DEBUGLOG("Failed to read file %s %s.\n",
 			 File, Exiv2::strError().c_str());
 		return NULL;
 	}
-	
+
+	try {
+		Image->readMetadata();
+	} catch (Exiv2::Error& e) {
+		DEBUGLOG("Failed to read EXIF metadata in %s.\n", File);
+		return NULL;
+	}
+
 	Exiv2::ExifData &ExifRead = Image->exifData();
 
 	// Read the tag out.
@@ -335,7 +326,7 @@ char* ReadGPSTimestamp(const char* File, char* DateStamp, char* TimeStamp, int*
 
 	// Copy the tag and return that.
 	char* Copy = strdup(Value.c_str());
-	
+
 	// Check if we have GPS tags.
 	Exiv2::Exifdatum& GPSData = ExifRead["Exif.GPSInfo.GPSVersionID"];
 
@@ -365,7 +356,7 @@ char* ReadGPSTimestamp(const char* File, char* DateStamp, char* TimeStamp, int*
 		RatNum3 = GPSData.toRational(2);
 		snprintf(TimeStamp, 12, "%02d:%02d:%02d",
 				RatNum1.first, RatNum2.first, RatNum3.first);
-		
+
 		GPSData = ExifRead["Exif.GPSInfo.GPSDateStamp"];
 		if (GPSData.count() < 3) {
 			*IncludesGPS = 0;
@@ -389,7 +380,6 @@ static void EraseGpsTags(Exiv2::ExifData &ExifInfo)
 {
 	// Search through, find the keys that we want, and wipe them
 	// Code below submitted by Marc Horowitz
-	Exiv2::ExifData::iterator Iter;
 	for (Exiv2::ExifData::iterator Iter = ExifInfo.begin();
 		Iter != ExifInfo.end(); )
 	{
@@ -480,7 +470,6 @@ int WriteGPSData(const char* File, const struct GPSPoint* Point,
 		DEBUGLOG("Failed to open file %s.\n", File);
 		return 0;
 	}
-	Image->readMetadata();
 	if (Image.get() == NULL)
 	{
 		// It failed if we got here.
@@ -488,7 +477,14 @@ int WriteGPSData(const char* File, const struct GPSPoint* Point,
 			 File, Exiv2::strError().c_str());
 		return 0;
 	}
-	
+
+	try {
+		Image->readMetadata();
+	} catch (Exiv2::Error& e) {
+		DEBUGLOG("Failed to read EXIF metadata in %s.\n", File);
+		return 0;
+	}
+
 	Exiv2::ExifData &ExifToWrite = Image->exifData();
 
 	// Make sure we're starting from a clean GPS IFD.
@@ -508,7 +504,7 @@ int WriteGPSData(const char* File, const struct GPSPoint* Point,
 	// Datum: the datum of the measured data. The default is WGS-84.
 	if (*Datum)
 		ExifToWrite["Exif.GPSInfo.GPSMapDatum"] = Datum;
-	
+
 	// Now start adding data.
 	// ALTITUDE.
 	// If no altitude was found in the GPX file, ElevDecimals will be -1
@@ -533,7 +529,7 @@ int WriteGPSData(const char* File, const struct GPSPoint* Point,
 		Value->read(ScratchBuf);
 		replace(ExifToWrite, Exiv2::ExifKey("Exif.GPSInfo.GPSAltitude"), Value.get());
 	}
-	
+
 	// LATITUDE
 	// Latitude reference: "N" or "S".
 	if (Point->Lat < 0)
@@ -578,7 +574,7 @@ int WriteGPSData(const char* File, const struct GPSPoint* Point,
 	}
 	Value->read(ScratchBuf);
 	replace(ExifToWrite, Exiv2::ExifKey("Exif.GPSInfo.GPSLatitude"), Value.get());
-	
+
 	// LONGITUDE
 	// Longitude reference: "E" or "W".
 	if (Point->Long < 0)
@@ -603,6 +599,17 @@ int WriteGPSData(const char* File, const struct GPSPoint* Point,
 	Value->read(ScratchBuf);
 	replace(ExifToWrite, Exiv2::ExifKey("Exif.GPSInfo.GPSLongitude"), Value.get());
 
+	// HDOP (Horizontal Dilution of Precision)
+	if (Point->HDOP >= 0) {
+		Value = Exiv2::Value::create(Exiv2::unsignedRational);
+		// Only store this with single decimal precision
+		ConvertToRational(Point->HDOP, 1, ScratchBuf, sizeof(ScratchBuf));
+		Value->read(ScratchBuf);
+		replace(ExifToWrite, Exiv2::ExifKey("Exif.GPSInfo.GPSDOP"), Value.get());
+		// This means GPSDOP holds HDOP, not PDOP
+		ExifToWrite["Exif.GPSInfo.GPSMeasureMode"] = "2";
+	}
+
 	// Heading.
 	if (Point->Heading >= 0) {
 		Value = Exiv2::Value::create(Exiv2::asciiString);
@@ -664,10 +671,9 @@ int WriteGPSData(const char* File, const struct GPSPoint* Point,
 	}
 
 	return 1;
-	
 }
 
-int WriteFixedDatestamp(const char* File, time_t Time)
+int WriteFixedDatestamp(const char* File, struct timespec Time)
 {
 	// Write the GPS data to the file...
 
@@ -684,7 +690,6 @@ int WriteFixedDatestamp(const char* File, time_t Time)
 		DEBUGLOG("Failed to open file %s.\n", File);
 		return 0;
 	}
-	Image->readMetadata();
 	if (Image.get() == NULL)
 	{
 		// It failed if we got here.
@@ -692,10 +697,17 @@ int WriteFixedDatestamp(const char* File, time_t Time)
 			 File, Exiv2::strError().c_str());
 		return 0;
 	}
-	
+
+	try {
+		Image->readMetadata();
+	} catch (Exiv2::Error& e) {
+		DEBUGLOG("Failed to read EXIF metadata in %s.\n", File);
+		return 0;
+	}
+
 	Exiv2::ExifData &ExifToWrite = Image->exifData();
-	
-	const struct tm TimeStamp = *gmtime(&Time);
+
+	const struct tm TimeStamp = *gmtime(&Time.tv_sec);
 	char ScratchBuf[100];
 
 	snprintf(ScratchBuf, sizeof(ScratchBuf), "%04d:%02d:%02d",
@@ -712,14 +724,14 @@ int WriteFixedDatestamp(const char* File, time_t Time)
 	Value->read(ScratchBuf);
 	ExifToWrite.erase(ExifToWrite.findKey(Exiv2::ExifKey("Exif.GPSInfo.GPSTimeStamp")));
 	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSTimeStamp"), Value.get());
-	
+
 	try {
 		Image->writeMetadata();
 	} catch (Exiv2::Error& e) {
 		DEBUGLOG("Failed to write to file %s.\n", File);
 		return 0;
 	}
-	
+
 	// Reset the mtime.
 	stat(File, &statbuf2);
 	utb.actime = statbuf2.st_atime;
@@ -739,14 +751,13 @@ int RemoveGPSExif(const char* File, int NoChangeMtime, int NoWriteExif)
 
 	// Open the file and start reading.
 	Exiv2::Image::UNIQUEPTR Image;
-	
+
 	try {
 		Image = Exiv2::ImageFactory::open(File);
 	} catch (Exiv2::Error& e) {
 		DEBUGLOG("Failed to open file %s.\n", File);
 		return 0;
 	}
-	Image->readMetadata();
 	if (Image.get() == NULL)
 	{
 		// It failed if we got here.
@@ -755,6 +766,13 @@ int RemoveGPSExif(const char* File, int NoChangeMtime, int NoWriteExif)
 		return 0;
 	}
 
+	try {
+		Image->readMetadata();
+	} catch (Exiv2::Error& e) {
+		DEBUGLOG("Failed to read EXIF metadata in %s.\n", File);
+		return 0;
+	}
+
 	Exiv2::ExifData &ExifInfo = Image->exifData();
 
 	if (ExifInfo.count() == 0)
@@ -764,7 +782,7 @@ int RemoveGPSExif(const char* File, int NoChangeMtime, int NoWriteExif)
 	}
 
 	EraseGpsTags(ExifInfo);
-	
+
 	if (!NoWriteExif) {
 		try {
 			Image->writeMetadata();
