@@ -6,7 +6,7 @@
  * Writes the output back into the GPS exif tags.
  */
 
-/* Copyright 2005-2023 Daniel Foote, Dan Fandrich.
+/* Copyright 2005-2024 Daniel Foote, Dan Fandrich.
  *
  * This file is part of gpscorrelate.
  *
@@ -74,6 +74,7 @@ static const struct option program_options[] = {
 	{ "heading", no_argument, 0, '('},
 	{ "direction", required_argument, 0, 'b'},
 	{ "max-heading", required_argument, 0, 'B'},
+	{ "no-photo-tz", no_argument, 0, ')'},
 	{ 0, 0, 0, 0 }
 };
 
@@ -102,14 +103,15 @@ static void PrintUsage(const char* ProgramName)
 	puts(  _("-B, --max-heading DEG    Max degrees between points that will be matched"));
 	puts(  _("-b, --direction DEG      Angular offset of camera from forward direction"));
 	puts(  _("-s, --show               Just show the GPS data from the given files"));
-	puts(  _("-o, --machine            Similar to --show but with machine-readable output"));
-	puts(  _("-x, --show-gpx           Similar to --show but with GPX output"));
+	puts(  _("-o, --machine            Like --show but with machine-readable CSV output"));
+	puts(  _("-x, --show-gpx           Like --show but with GPX output"));
 	puts(  _("-r, --remove             Strip GPS tags from the given files"));
 	puts(  _("-t, --ignore-tracksegs   Interpolate between track segments, too"));
 	puts(  _("-M, --no-mtime           Don't change mtime of modified files"));
 	puts(  _("-f, --fix-datestamps     Fix broken GPS datestamps written with ver. < 1.5.2"));
 	puts(  _("    --degmins            Write location as DD MM.MM (was default before v1.5.3)"));
 	puts(  _("-O, --photooffset SECS   Offset added to photo time to make it match the GPS"));
+	puts(  _("    --no-photo-tz        Ignore time zone tags within photos"));
 	puts(  _("-h, --help               Display this help message"));
 	puts(  _("-v, --verbose            Show more detailed output"));
 	puts(  _("-V, --version            Display version information"));
@@ -154,7 +156,8 @@ static int ShowFileDetails(const char* File, enum OutputFormat Format,
 	int IncludesGPS = 0;
 	Lat = Long = 0;
 	Elev = NAN; /* Elevation is optional, so this means it's missing */
-	char* Time = ReadExifData(File, &Lat, &Long, &Elev, &IncludesGPS);
+	long OffsetTime = 0;
+	char* Time = ReadExifData(File, &IncludesGPS, &Lat, &Long, &Elev, &OffsetTime);
 	int rc = 1;
 	char* OldLocale = NULL;
 	static int Started = 0;
@@ -177,8 +180,6 @@ static int ShowFileDetails(const char* File, enum OutputFormat Format,
 			PACKAGE_VERSION);
 		if (Options->AutoTimeZone && Time)
 		{
-			/* Use the local time zone as of the date of first picture
-			 * as the time for correlating all the remainder. */
 			SetAutoTimeZoneOptions(Time, Options);
 			Options->AutoTimeZone = 0;
 		}
@@ -209,14 +210,15 @@ static int ShowFileDetails(const char* File, enum OutputFormat Format,
 			} else if (Format == GPX_FORMAT)
 			{
 				/* Now convert the time into Unixtime. */
-				time_t PhotoTime = ConvertTimeToUnixTime(Time, EXIF_DATE_FORMAT, Options);
+				struct timespec PhotoTime = ConvertTimeToUnixTime(Time, EXIF_DATE_FORMAT,
+						OffsetTime, NULL, Options);
 				static time_t LastGpxTime;
-				if (PhotoTime < LastGpxTime)
+				if (CompareTimes(PhotoTime, LastGpxTime) < 0)
 					fprintf(stderr, _("Warning: image files are not ordered by time.\n"));
-				LastGpxTime = PhotoTime;
+				LastGpxTime = PhotoTime.tv_sec;
 
 				char GpxTime[24];
-				struct tm* Tm = gmtime(&PhotoTime);
+				struct tm* Tm = gmtime(&PhotoTime.tv_sec);
 				strftime(GpxTime, sizeof(GpxTime), "%Y-%m-%dT%H:%M:%SZ", Tm);
 
 				char *SafeFile = strdup(File);
@@ -284,7 +286,7 @@ static void ShowFileDone(enum OutputFormat Format)
 			   "</gpx>\n");
 	}
 }
-			
+
 /* Remove all GPS exif tags from a file. Not really that useful, but... */
 static int RemoveGPSTags(const char* File, int NoChangeMtime, int NoWriteExif)
 {
@@ -321,14 +323,14 @@ static int FixDatestamp(const char* File, int AdjustmentHours, int AdjustmentMin
 		rc = 0;
 	} else {
 		/* Check the timestamp. */
-		time_t PhotoTime = ConvertToUnixTime(OriginalDateStamp, EXIF_DATE_FORMAT,
-			AdjustmentHours, AdjustmentMinutes);
-		
+		struct timespec PhotoTime = ConvertToUnixTime(OriginalDateStamp, EXIF_DATE_FORMAT,
+			AdjustmentHours * 3600 + AdjustmentMinutes * 60);
+
 		snprintf(CombinedTime, sizeof(CombinedTime), "%s %s", DateStamp, TimeStamp);
 
-		time_t GPSTime = ConvertToUnixTime(CombinedTime, EXIF_DATE_FORMAT, 0, 0);
+		struct timespec GPSTime = ConvertToUnixTime(CombinedTime, EXIF_DATE_FORMAT, 0);
 
-		if (PhotoTime != GPSTime) {
+		if (CompareTimes(PhotoTime, GPSTime.tv_sec)) {
 			/* Timestamp is wrong. Fix it.
 			 * Should be photo time - this also corrects
 			 * GPSTimestamp, which was wrong too. */
@@ -340,9 +342,9 @@ static int FixDatestamp(const char* File, int AdjustmentHours, int AdjustmentMin
 			char GPSTimeFormat[100];
 
 			strftime(PhotoTimeFormat, sizeof(PhotoTimeFormat),
-				 "%a %b %d %H:%M:%S %Y UTC", gmtime(&PhotoTime));
+				 "%a %b %d %H:%M:%S %Y UTC", gmtime(&PhotoTime.tv_sec));
 			strftime(GPSTimeFormat, sizeof(GPSTimeFormat),
-				 "%a %b %d %H:%M:%S %Y UTC", gmtime(&GPSTime));
+				 "%a %b %d %H:%M:%S %Y UTC", gmtime(&GPSTime.tv_sec));
 			printf(_("%s: Wrong timestamp:\n   Photo:     %s\n"
 				 "   GPS:       %s\n   Corrected: %s\n"),
 					File, PhotoTimeFormat, GPSTimeFormat, PhotoTimeFormat);
@@ -375,7 +377,7 @@ int main(int argc, char** argv)
 	/* Parse our command line options. */
 	/* But first, some variables to store stuff in. */
 	int c;
-	
+
 	struct GPSTrack* Track = NULL;/* Array of lists of GPS waypoints. The
 					 final entry of all 0 signals the end. */
 	int NumTracks = 0;	     /* Number of track structures at Track,
@@ -394,9 +396,10 @@ int main(int argc, char** argv)
 	int RemoveTags = 0;
 	int DoBetweenTrackSegs = 0;
 	int NoChangeMtime = 0;
+	int NoPhotoTz = 0;
 	int FixDatestamps = 0;
 	int DegMinSecs = 1;
-	int PhotoOffset = 0;
+	double PhotoOffset = 0;
 	int WriteHeading = 0;
 	int HeadingOffset = -1;      /* Negative means disabled */
 	int MaxHeadingDelta = -1;    /* Negative means disabled */
@@ -479,7 +482,8 @@ int main(int argc, char** argv)
 					{
 						/* Found colon. Split into two. */
 						sscanf(optarg, "%d:%d", &TimeZoneHours, &TimeZoneMins);
-						if (TimeZoneHours < 0)
+						/* Can't look at sign of TimeZoneHours because it might be -0 */
+						if (*optarg == '-')
 						    TimeZoneMins *= -1;
 					} else {
 						/* No colon. Just parse. */
@@ -491,7 +495,7 @@ int main(int argc, char** argv)
 			case 'O':
 				if (optarg)
 				{
-					PhotoOffset = atoi(optarg);
+					PhotoOffset = atof(optarg);
 				}
 				break;
 			case 'i':
@@ -529,6 +533,10 @@ int main(int argc, char** argv)
 				/* This option specifies to write the GPSTrack tag, if possible. */
 				WriteHeading = 1;
 				break;
+			case ')':
+				/* This option causes OffsetTime tags within photos to be ignored. */
+				NoPhotoTz = 1;
+				break;
 			case 'b':
 				/* This option gives us the offset of the camera from forward. */
 				if (optarg)
@@ -658,8 +666,9 @@ int main(int argc, char** argv)
 	struct CorrelateOptions Options;
 	Options.NoWriteExif   = NoWriteExif;
 	Options.OverwriteExisting = OverwriteExisting;
-	Options.NoInterpolate = (Interpolate ? 0 : 1);
+	Options.NoInterpolate = !Interpolate;
 	Options.AutoTimeZone  = !HaveTimeAdjustment;
+	Options.TimeZoneFromPhoto = !NoPhotoTz && !HaveTimeAdjustment;
 	Options.TimeZoneHours = TimeZoneHours;
 	Options.TimeZoneMins  = TimeZoneMins;
 	Options.FeatherTime   = FeatherTime;
@@ -706,7 +715,7 @@ int main(int argc, char** argv)
 			fprintf(stderr, _("A time offset must be given with the -z option to fix photos.\n"));
 			exit(EXIT_FAILURE);
 		}
-		
+
 		int result = 1;
 		while (optind < argc)
 		{
@@ -736,7 +745,7 @@ int main(int argc, char** argv)
 	/* Make it all look nice and pretty... so the user knows what's going on. */
 	printf(_("\nCorrelate: "));
 	if (ShowDetails) printf("\n");
-	
+
 	/* A few variables that we'll require later. */
 	struct GPSPoint* Result;
 	char* File;
@@ -749,6 +758,8 @@ int main(int argc, char** argv)
 	int TooFar     = 0;
 	int NoDate     = 0;
 	int GPSPresent = 0;
+	long UsedOffset = NO_OFFSET_TIME;
+	int OffsetCount = 0;
 
 	/* Now it is time to correlate the photos. Feed one in at a time, and
 	 * see what happens.*/
@@ -758,10 +769,16 @@ int main(int argc, char** argv)
 	while (optind < argc)
 	{
 		File = argv[optind++];
+		long ThisOffset;
 		/* Pass the file along to Correlate and see what happens. */
-		Result = CorrelatePhoto(File, &Options);
+		Result = CorrelatePhoto(File, &ThisOffset, &Options);
+		if (UsedOffset == NO_OFFSET_TIME)
+			UsedOffset = ThisOffset;
+		else if (ThisOffset != NO_OFFSET_TIME && ThisOffset != UsedOffset)
+			/* Multiple offsets were encountered */
+			++OffsetCount;
 
-		/* Was result NULL? */
+		/* Was result non-NULL? */
 		if (Result)
 		{
 			/* Result not null. But what did happen? */
@@ -774,8 +791,7 @@ int main(int argc, char** argv)
 				} else {
 					printf(".");
 				}
-			}
-			if (Options.Result == CORR_INTERPOLATED)
+			} else if (Options.Result == CORR_INTERPOLATED)
 			{
 				MatchInter++;
 				if (ShowDetails)
@@ -784,8 +800,7 @@ int main(int argc, char** argv)
 				} else {
 					printf("/");
 				}
-			}
-			if (Options.Result == CORR_ROUND)
+			} else if (Options.Result == CORR_ROUND)
 			{
 				MatchRound++;
 				if (ShowDetails)
@@ -794,8 +809,7 @@ int main(int argc, char** argv)
 				} else {
 					printf("<");
 				}
-			}
-			if (Options.Result == CORR_EXIFWRITEFAIL)
+			} else if (Options.Result == CORR_EXIFWRITEFAIL)
 			{
 				WriteFail++;
 				if (ShowDetails)
@@ -828,8 +842,7 @@ int main(int argc, char** argv)
 				} else {
 					printf("-");
 				}
-			}
-			if (Options.Result == CORR_TOOFAR)
+			} else if (Options.Result == CORR_TOOFAR)
 			{
 				TooFar++;
 				if (ShowDetails)
@@ -838,8 +851,7 @@ int main(int argc, char** argv)
 				} else {
 					printf("^");
 				}
-			}
-			if (Options.Result == CORR_NOEXIFINPUT)
+			} else if (Options.Result == CORR_NOEXIFINPUT)
 			{
 				NoDate++;
 				if (ShowDetails)
@@ -848,8 +860,7 @@ int main(int argc, char** argv)
 				} else {
 					printf("?");
 				}
-			}
-			if (Options.Result == CORR_GPSDATAEXISTS)
+			} else if (Options.Result == CORR_GPSDATAEXISTS)
 			{
 				GPSPresent++;
 				if (ShowDetails)
@@ -867,9 +878,9 @@ int main(int argc, char** argv)
 
 		/* And, once we've got here, we've finished with that file.
 		 * We can now do the next one. Now wasn't that too easy? */
-		
+
 	} /* End while parse command line files. */
-	
+
 	/* Right, so now we're done. That really wasn't that hard. Right? */
 
 	/* Add a new line if we were doing the not-show-details thing. */
@@ -880,12 +891,16 @@ int main(int argc, char** argv)
 
 	/* Print details of what happened. */
 	printf(_("\nCompleted correlation process.\n"));
-	if (ShowDetails)
+	if (ShowDetails && UsedOffset != NO_OFFSET_TIME) {
 		/* This has to be shown at the end in case auto time zone
 		 * was used, since it isn't known before the first file
 		 * is processed. */
-		printf(_("Used time zone offset %d:%02d\n"),
-		       Options.TimeZoneHours, abs(Options.TimeZoneMins));
+		if (OffsetCount)
+			printf(_("Used time zone offset (multiple)\n"));
+		else
+			printf(_("Used time zone offset %s%ld:%02ld\n"), UsedOffset < 0 ? "-" : "",
+				   labs(UsedOffset / 3600), labs(UsedOffset % 3600) / 60);
+	}
 	printf(_("Matched: %5d (%d Exact, %d Interpolated, %d Rounded).\n"),
 			MatchExact + MatchInter + MatchRound,
 			MatchExact, MatchInter, MatchRound);
@@ -904,7 +919,7 @@ int main(int argc, char** argv)
 	}
 	free(Track);
 	free(Datum);
-	
+
 	if (WriteFail)
 		/* A write failure is considered serious */
 		return EXIT_FAILURE;
