@@ -5,7 +5,7 @@
  * the Unix time functions.
  */
 
-/* Copyright 2005-2012 Daniel Foote, Dan Fandrich.
+/* Copyright 2005-2024 Daniel Foote, Dan Fandrich.
  *
  * This file is part of gpscorrelate.
  *
@@ -25,10 +25,12 @@
  */
 
 #include <sys/types.h>
+#include <time.h>
 
-#define EXIF_DATE_FORMAT "%d:%d:%d %d:%d:%d"
+// The EXIF date/time format extended to subsecond resolution
+#define EXIF_DATE_FORMAT "%d:%d:%d %d:%d:%d.%9[0-9]"
 #define GPX_DATE_FORMAT "%d-%d-%dT%d:%d:%dZ"
 
-time_t ConvertToUnixTime(const char* StringTime, const char* Format,
-		int TZOffsetHours, int TZOffsetMinutes);
-
+struct timespec ConvertToUnixTime(const char* StringTime, const char* Format,
+		long TZOffset);
+int CompareTimes(struct timespec a, time_t b);
